#include "base.h"
#include "NetLib.h"
#include "zlib.h"

#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>

static const uint32 KMessageDuplicates = 3;
static const uint32 KInitialDelay = 30;
static const uint32 KTimerInterval = 30;
static const uint32 KArenaKillInterval = 250;

struct Peer
{
	struct Arena
	{
		uint32 iId;
		std::string iName;
		StringSet iPlayers;

		uint32 iLastUpdateTime;
	};
	typedef std::list <Arena> Arenas;

	InetAddress iAddress;
	uint32 iPasswordHash;
	bool iSendOnly;
	bool iSendPlayerList;
	bool iSendMessages;
	bool iReceiveMessages;
	StringSequence iForwards;

	uint16 iPopulation;
	uint32 iTimestamps[0x100];
	Arenas iArenas;
};

extern "C" {
#include "asss.h"
#include "redirect.h"
#include "peer/peer.h"
}

struct PeerPacket
{
	uint8 t1;
	uint8 t2;
	uint32 password;
	uint8 t3;
	uint8 type;
	uint32 timestamp;
};

local Imodman *mm;
local Inet *net;
local Iconfig *cfg;
local Ilogman *lm;
local Imainloop *ml;
local Ichat *chat;
local Iarenaman *arenaman;
local Iplayerdata *pd;
local Iredirect *redirect;
local ConfigHandle cfgHandle;
local pthread_mutex_t lock;

typedef DataWriter <LittleEndian, AsciiZ> writer_t;
typedef DataReader <LittleEndian, AsciiZ> reader_t;
typedef std::list <Peer> PeerList;
PeerList gPeerList;

local int PeriodicUpdate(void *)
{
	Arena *a;
	Player *p;
	Link *link;
	Iarenaman *aman = arenaman;
	int total;

	ticks_t now = current_ticks();

	writer_t arenaList;
	aman->Lock();
	aman->GetPopulationSummary(&total, NULL);
	pd->Lock();
	FOR_EACH_ARENA(a)
	{
		arenaList.AppendUint32(0);
		arenaList.AppendString(a->name);

		Link *link;
		FOR_EACH_PLAYER(p)
			if(p->arena == a)
				arenaList.AppendString(p->name);

		arenaList.AppendString("");
	}
	pd->Unlock();
	aman->Unlock();

	pthread_mutex_lock(&lock);
	for(PeerList::iterator i = gPeerList.begin(); i != gPeerList.end(); ++i)
	{
		for(Peer::Arenas::iterator j = i->iArenas.begin(); j != i->iArenas.end(); )
		{
			Peer::Arenas::iterator here = j++;
			if(TICK_DIFF(now, here->iLastUpdateTime) >= KArenaKillInterval)
				i->iArenas.erase(here);
		}

		writer_t writer;
		writer.AppendUint8(0x00);
		writer.AppendUint8(0x01);
		writer.AppendUint32(i->iPasswordHash);
		writer.AppendUint8(0xFF);
		writer.AppendUint8((i->iSendPlayerList) ? 0x01 : 0x04);
		writer.AppendUint32(now);

		if(i->iSendPlayerList)
			writer.AppendString(arenaList.AsBString());
		else
			writer.AppendUint16(total);

		sockaddr_in sa = i->iAddress.AsSockaddr();
		net->ReallyRawSend(&sa, const_cast <byte *> (writer.AsBString().Data()), writer.Length(), 0);
	}
	pthread_mutex_unlock(&lock);

	return TRUE;
}

local void ChatMessage(LinkedList *set, char type, char sound, Player *from, const char *message)
{
	if(type != MSG_ZONE && type != MSG_MODCHAT)
		return;

	pthread_mutex_lock(&lock);

	for(PeerList::const_iterator i = gPeerList.begin(); i != gPeerList.end(); ++i)
	{
		if(!i->iSendMessages)
			continue;

		writer_t writer;
		writer.AppendUint8(0x00);
		writer.AppendUint8(0x01);
		writer.AppendUint32(i->iPasswordHash);
		writer.AppendUint8(0xFF);
		writer.AppendUint8((type == MSG_ZONE) ? 0x02 : 0x03);
		writer.AppendUint32(current_ticks());
		writer.AppendUint8(0x00);
		writer.AppendString(message);

		sockaddr_in sa = i->iAddress.AsSockaddr();

		//
		// Make sure the datagram gets there.
		//
		for(uint32 i = 0; i < KMessageDuplicates; ++i)
			net->ReallyRawSend(&sa, const_cast <byte *> (writer.AsBString().Data()), writer.Length(), 0);
	}

	pthread_mutex_unlock(&lock);
}

local int ArenaRequest(Player *p, int arenatype, const char *arena)
{
	pthread_mutex_lock(&lock);

	for(PeerList::const_iterator i = gPeerList.begin(); i != gPeerList.end(); ++i)
		for(StringSequence::const_iterator j = i->iForwards.begin(); j != i->iForwards.end(); ++j)
			if(!strcasecmp(arena, j->c_str()))
			{
				pthread_mutex_unlock(&lock);

				Target t;
				t.type = Target::T_PLAYER;
				t.u.p = p;

				return redirect->RawRedirect(&t, i->iAddress.GetIpAddress().c_str(), i->iAddress.GetPort(), arenatype, arena);
			}

	pthread_mutex_unlock(&lock);
	return FALSE;
}

local int GetPopulationSummary()
{
	pthread_mutex_lock(&lock);

	int ret = 0;
	for(PeerList::const_iterator i = gPeerList.begin(); i != gPeerList.end(); ++i)
		if(i->iArenas.empty())
			ret += i->iPopulation;
		else
			for(Peer::Arenas::const_iterator j = i->iArenas.begin(); j != i->iArenas.end(); ++j)
				ret += j->iPlayers.size();
	pthread_mutex_unlock(&lock);
	return ret;
}

local int AppendArenasToPacket(void *pkt, int maxLength, int showHidden)
{
	if(!pkt || maxLength <= 3)
		return 0;

	pthread_mutex_lock(&lock);
	writer_t writer;

	for(PeerList::const_iterator p = gPeerList.begin(); p != gPeerList.end(); ++p)
		for(Peer::Arenas::const_iterator a = p->iArenas.begin(); a != p->iArenas.end(); ++a)
			if(a->iName.length() + 3 + writer.Length() < maxLength && (showHidden || a->iName[0] != '#'))
			{
				writer.AppendString(a->iName);
				writer.AppendUint16(a->iPlayers.size());
			}
	pthread_mutex_unlock(&lock);

	BString data = writer.AsBString();

	memcpy(pkt, data.Data(), data.GetLength());
	return data.GetLength();
}

local int FindPlayer(const char *target, int score, char **name, char **arena)
{
	if(!target || !name || !arena)
		return score;

	pthread_mutex_lock(&lock);

	for(PeerList::const_iterator p = gPeerList.begin(); p != gPeerList.end(); ++p)
		for(Peer::Arenas::const_iterator a = p->iArenas.begin(); a != p->iArenas.end(); ++a)
			for(StringSet::const_iterator i = a->iPlayers.begin(); i != a->iPlayers.end(); ++i)
			{
				const char *pos = strcasestr(i->c_str(), target);
				if(pos && pos-target < score)
				{
					if(!*name)
						afree(*name);
					if(!*arena)
						afree(*arena);

					*name = astrdup(i->c_str());
					*arena = astrdup(a->iName.c_str());
					score = pos-target;
				}
			}

	pthread_mutex_unlock(&lock);
	return score;
}

local void HandlePlayerList(Peer & peer, const BString & packet)
{
	peer.iArenas.clear();

	uint32 now = current_ticks();

	reader_t reader(packet);
	while(reader.BytesRemaining())
	{
		uint32 id = reader.ReadUint32();
		std::string name = reader.ReadString();

		StringSet * playerList = 0;
		for(Peer::Arenas::iterator i = peer.iArenas.begin(); i != peer.iArenas.end(); ++i)
		{
			if(i->iName == name)
			{
				i->iLastUpdateTime = now;
				playerList = &i->iPlayers;
				break;
			}
		}

		if(playerList == 0)
		{
			Peer::Arena arena;
			arena.iId = id;
			arena.iName = name;
			arena.iLastUpdateTime = now;
			peer.iArenas.push_back(arena);
			playerList = &peer.iArenas.back().iPlayers;
		}

		playerList->clear();
		for(std::string playerName = reader.ReadString(); !playerName.empty() && reader.BytesRemaining(); playerName = reader.ReadString())
			playerList->insert(playerName);
	}
}

local void HandleChatMessage(Peer & peer, const BString & packet)
{
	if(!peer.iReceiveMessages)
		return;

	reader_t reader(packet);

	u8 type = reader.ReadUint8();
	std::string message = reader.ReadString();

	chat->SendZoneMessageNoForward("%s", message.c_str());
}

local void HandleOpMessage(Peer & peer, const BString & packet)
{
	if(!peer.iReceiveMessages)
		return;

	reader_t reader(packet);

	u8 type = reader.ReadUint8();
	std::string message = reader.ReadString();

	chat->SendModMessageNoForward("%s", message.c_str());
}

local void PeerUpdate(struct sockaddr_in *sin, byte *pkt, int len, void *v)
{
	if(len < 12 || pkt[0] != 0x00 || pkt[1] != 0x01 || pkt[6] != 0xFF)
		return;

	pthread_mutex_lock(&lock);
	PeerPacket *packet = (PeerPacket *)pkt;
	for(PeerList::iterator p = gPeerList.begin(); p != gPeerList.end(); ++p)
	{
		if(p->iSendOnly ||
		   p->iAddress.AsUint32HostOrder() != ntohl(sin->sin_addr.s_addr) || p->iAddress.GetPort() != ntohs(sin->sin_port) ||
		   p->iPasswordHash != packet->password ||
		   p->iTimestamps[packet->timestamp & 0xFF] == packet->timestamp)
			continue;

		p->iTimestamps[packet->timestamp & 0xFF] = packet->timestamp;
		switch(packet->type)
		{
			case 1:
				HandlePlayerList(*p, BString(pkt + sizeof(PeerPacket), len - sizeof(PeerPacket)));
				break;
			case 2:
				HandleChatMessage(*p, BString(pkt + sizeof(PeerPacket), len - sizeof(PeerPacket)));
				break;
			case 3:
				HandleOpMessage(*p, BString(pkt + sizeof(PeerPacket), len - sizeof(PeerPacket)));
				break;
			case 4:
				p->iPopulation = *(u16 *)(pkt + sizeof(PeerPacket));
				break;
			default:
				break;
		}
		break;
	}
	pthread_mutex_unlock(&lock);
}

local void ConfigChanged(void * clos)
{
	ml->ClearTimer(PeriodicUpdate, NULL);

	pthread_mutex_lock(&lock);
	gPeerList.clear();

	bool startupTimer = false;
	for(int i = 0; i < 8; ++i)
	{
		try
		{
			char peerSection[8];
			sprintf(peerSection, "Peer%u", i);

			const char *addr = cfg->GetStr(cfgHandle, peerSection, "Address");
			if(!addr)
				continue;

			u32 hash = 0xDEADBEEF;
			const char *password = cfg->GetStr(cfgHandle, peerSection, "Password");
			if(password)
				hash = ~crc32(0, (const Bytef *)password, strlen(password));

			const char *forwards = cfg->GetStr(cfgHandle, peerSection, "Arenas");
			if(!forwards)
				forwards = "";

			Peer p;
			p.iAddress = InetAddress(addr);
			p.iPasswordHash = hash;
			p.iSendOnly = cfg->GetInt(cfgHandle, peerSection, "SendOnly", 0) != 0;
			p.iSendPlayerList = cfg->GetInt(cfgHandle, peerSection, "SendPlayerList", 0) != 0;
			p.iSendMessages = cfg->GetInt(cfgHandle, peerSection, "SendMessages", 1) != 0;
			p.iReceiveMessages = cfg->GetInt(cfgHandle, peerSection, "ReceiveMessages", 1) != 0;
			p.iForwards = String::Split(forwards, ',');
			p.iPopulation = 0;

			for(StringSequence::iterator j = p.iForwards.begin(); j != p.iForwards.end(); ++j)
				String::Trim(*j);

			gPeerList.push_back(p);

			startupTimer = true;
		}
		catch(const malformed_error &)
		{
		}
	}
	pthread_mutex_unlock(&lock);

	if(startupTimer)
		ml->SetTimer(PeriodicUpdate, KInitialDelay, KTimerInterval, NULL, NULL);
}

local Ipeer _int =
{
	INTERFACE_HEAD_INIT(I_PEER, "peer")
	GetPopulationSummary,
	AppendArenasToPacket,
	FindPlayer,
	ArenaRequest
};

extern "C" EXPORT int MM_peer(int action, Imodman *mm_, Arena *arena)
{
	mm = mm_;

	switch(action)
	{
		case MM_LOAD:
			lm = (Ilogman *)mm->GetInterface(I_LOGMAN, ALLARENAS);
			ml = (Imainloop *)mm->GetInterface(I_MAINLOOP, ALLARENAS);
			arenaman = (Iarenaman *)mm->GetInterface(I_ARENAMAN, ALLARENAS);
			pd = (Iplayerdata *)mm->GetInterface(I_PLAYERDATA, ALLARENAS);
			net = (Inet *)mm->GetInterface(I_NET, ALLARENAS);
			cfg = (Iconfig *)mm->GetInterface(I_CONFIG, ALLARENAS);
			chat = (Ichat *)mm->GetInterface(I_CHAT, ALLARENAS);
			redirect = (Iredirect *)mm->GetInterface(I_REDIRECT, ALLARENAS);

			if(!lm || !ml || !arenaman || !pd || !net || !cfg || !chat || !redirect)
				return MM_FAIL;

			mm->RegCallback(CB_CONNINIT, (void *)PeerUpdate, ALLARENAS);
			mm->RegCallback(CB_SENDCHATMSG, (void *)ChatMessage, ALLARENAS);
			mm->RegInterface(&_int, ALLARENAS);
			cfgHandle = cfg->OpenConfigFile(NULL, "peer.conf", ConfigChanged, NULL);

			pthread_mutex_init(&lock, NULL);
			gPeerList = PeerList();
			ConfigChanged(NULL);
			break;

		case MM_UNLOAD:
			if(!mm->UnregInterface(&_int, ALLARENAS))
				return MM_FAIL;

			ml->ClearTimer(PeriodicUpdate, NULL);
			pthread_mutex_destroy(&lock);

			cfg->CloseConfigFile(cfgHandle);
			mm->UnregCallback(CB_SENDCHATMSG, (void *)ChatMessage, ALLARENAS);
			mm->UnregCallback(CB_CONNINIT, (void *)PeerUpdate, ALLARENAS);
			mm->ReleaseInterface(redirect);
			mm->ReleaseInterface(chat);
			mm->ReleaseInterface(cfg);
			mm->ReleaseInterface(net);
			mm->ReleaseInterface(pd);
			mm->ReleaseInterface(arenaman);
			mm->ReleaseInterface(ml);
			mm->ReleaseInterface(lm);
			break;

		default:
			return MM_FAIL;
	}

	return MM_OK;
}
